/*
 * This program tries to locate the light-beams falling on the Hawaii Sensor.
 * It locates the weighted center of the "star-matter", and then prints the coordinates, and some summary stats.
 * The input is a 1024 x 1024 pixel 32-bit (monochrome) tiff file, usually from cam2tiff.

 * Copyright (C) Richard Neill 2011, <cam2tiff at REMOVE.ME.richardneill.org>. This program is Free Software. You can
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
 * For the details, please see: http://www.gnu.org/licenses/gpl.html

 * Thanks to Edmund Jackson for assistance with the algorithm
 */

/* TODO:
 * 1. Consider adding a threshold to cut-off noise, i.e. a minimum brightness below which beams cannot be found. Should this be a constant, a parameter,
 * or a fraction of the peak brightness? Would it actually be counterproductive?
 *
 * 2. The heuristic for VALID_MINRELPXBRI doesn't work very well if there is a high noise floor. It's deirable to keep VALID_MINRELPXBRI low (0.005) with a normal
 *    high dynamic range + low noise floor. But if the noise floor is too high, and the dynamic range is low, this test identifies valid beams as invalid.
 *    Improvement: consider a beam valid if its central pixels is at least 0.5% as bright as the brightest star AND it is above the average-image noise-floor.
 *
 * 3. NB take care to avoid star splitting that prevents the (below-threshold) faint ring around star n being detected much later as a less-bright star m,
 *    while still ensuring that nearby double-stars (with an intensity-dip between them) don't get merged, The find_adjacent_star() approach seems to work
 *    ok, but it might be possible to do even better.
 */


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <tiffio.h>
#include <math.h>
#include <inttypes.h>

/* Finder Algorithm */
#define MAX_STARS		32		/* Stop after finding at most this many beams. MAX value is 32 i.e. bits in an int. */
#define LUM_THRESHOLD		0.03		/* Fractional threshold for star "growth" wrt the brightest pixel in that star. Too high means we don't collect all the star pixels (and may make stars split); too low picks up noise, whiteout. 3% is a good value */
#define MAX_PIXELS_PERSTAR	225		/* Max number of pixels that a beam can cover (makes algorithm terminate as "whiteout"). [Also, grow_star() is O2 wrt this.] */

/* Validity Heuristic */
#define VALID_MAXPIXELCOUNT	25		/* Should a particular beam be considered "valid"? Max number of pixels in a beam. Should be < MAX_PIXELS_PERSTAR */
#define VALID_FOCUSMIN		0.10		/* Expect at least 15% of beam to be focussed rightly, dead on the central pixel. (this is the cutoff for rejecting a beam; be generous) */
#define VALID_MINRELPXBRI	0.005		/* Expect this beam's peak brightness to be  > 0.005 *  the peak pixel brightness of the entire image */

/* Hawaii properties */
#define	WIDTH			1024		/* Expected width and height for Hawaii chip (full-frame) */
#define	HEIGHT			1024
#define	QUADS			4
#define LUM_SCALEFACTOR_8BIT	256		/* The calculated luminosity is scaled down by this, to match the units of the .cam file (MSB = bit-24) and Hawaii-ADC, whereas the TIFF has MSB = bit32. */

/* TIFF properties */
#define	TIFF_EXPECT_BPS		32		/* Expected tiff properties, matching cam2tiff. TIFF_EXPECT_BPS must match uint32 for pixels[][] etc */
#define TIFF_EXPECT_SPP		1
#define TIFF_EXPECT_RPS		1
#define TIFF_EXPECT_ORIENT  	ORIENTATION_TOPLEFT

#define TIFF_OUTPUT_BPS 	16		/* Config for output file (red overlay), Matches output_r[][] etc */
#define TIFF_OUTPUT_SPP 	3

#define	TIFF_EXTN		"tiff"		/* File extensions */
#define	TIF_EXTN		"tif"

/* Misc */
#define DEBUG 			0		/* Turn on debugging? */
#define DISPLAY_PROGRAM		"display"	/* The image viewer. Use without path, to support /usr/local/bin/display if needed. */

#if DEBUG == 1
	#define debug_printf(...)   fprintf(stderr, __VA_ARGS__)
#else
	#define debug_printf(...)
#endif


uint32  pixels [HEIGHT][WIDTH];		/* Array for raw pixel data. Warning: this is  pixels[y][x]  (not [x][y]), to match the tiff. MUST match TIFF_EXPECT_BPS */
uint32  star_map [HEIGHT][WIDTH];	/* Array of pixels, star-number (if any) is identified by bitfield. (0 means "Not in a star", 0x8 means "detected in star 3". At most 1 bit set.) */

uint16  output_r[HEIGHT][WIDTH];	/* Array for output file. 16-bit per channel, matches TIFF_OUTPUT_BPS */
uint16  output_g[HEIGHT][WIDTH];
uint16  output_b[HEIGHT][WIDTH];

uint32  peak_for_normalising = 0;	/* The Brightest pixel-value in the image (used for normalising the plot) */
int     stars_found = 0;		/* Stars found so far, also used (via star_n and sbit() ) as the bitfield-position in star_map[][]. */
double  lum_threshold = LUM_THRESHOLD;  /* Default value, may be overridden */


typedef struct {			/* Struct for a detected star. */
	int n;				/* detected number. 0 first, should be (but not always) the brightest. */
	int x, y;			/* integer x and y coordinates of the brightest pixel  (in X11-coordinates) */
	uint32 brightest;		/* pixel value of the brightest pixel*/
	double c_x, c_y;		/* fractional x and y coordinates of the centroid  (i.e. "Center of Mass")   (in X11-coordinates) */
	int h_q, h_x, h_y; 		/* Hawaii coordinates: quadrant and x/y within pixel. */
	int pixels;			/* Number of pixels making up the star */
	long long total_lum;		/* Total luminosity of star. */
	float avg_lum;			/* Average luminosity (= total_lum / pixels) */
	float  focus;			/* Central focus. (The fraction of the total luminosity that is in the (integer) pixel identified as "the" center). */
	char *valid;			/* Is the beam considered, heuristially, to be valid? "yes"/"no" */
} star;

star stars[MAX_STARS];			/* Resulting stars! */

void print_help(char *argv0){
	fprintf (stderr, "Beamfinder: search a TIFF file to find the positions and strengths of the light-beams falling on the sensor.\n"
			 "This is used to align (and check) the beam positions on the sensor, and to select which pixels to read out.\n"
			 "The algorithm stops after finding at most %d beams, and 'grows' each beam to %d pixels at most.\n\n"
			 "  USAGE:     %s   [OPTS]  infile.%s\n"
			 "  OPTIONS:   -h                 show help\n"
			 "             -p                 plot the file: display a normalised image, with the detected beam positions in red\n"
			 "             -t   threshold     adjust luminosity-threshold for pixel/peak_pixel detection. (default: %4.2f).\n"
			 "                                higher values (0.2 - 1.0) improve noise-immunity, but discard faint pixels as noise.\n"
			 "\n"
			 "  OUTPUT:    This is designed for ease of parsing. Stdout/stderr are separated. Headers begin with '#'.\n"
			 "             Stars are listed in descending order [of brightest pixel], up to at most %d. The columns are:\n"
			 "             * Hawaii_Coordinates (Quad,X,Y):  the (integer) positions of the centroidal pixel in each beam.\n"
			 "             * Beam Pixels:                    number of pixels bright enough to be considered \"in\" the beam.\n"
			 "             * Focus & Alignment:              fraction of the beam that lies in the centroidal pixel. Ideally 1.0\n"
			 "             * Valid Beam?:                    heuristic: is this detected beam (plausibly) good? yes/no.\n"
			 "             * Beam Total Luminosity:          sum of the pixel-values in the beam, scaled (/%d) to the .cam units\n"
			 "             * Average Luminosity:             calculated as: total luminosity / pixel count.\n"
			 "             * Centroid Position (X,Y):        positions of the beam centroid, floats, in X11-coordinate system.\n"
			 "\n"
			 "Note 1: A well-aligned, and sharply focused system should have a Focus & Alignment score of nearly 1.0. This means\n"
			 "that all of the beam illumination falls within the size of a single-pixel, AND the beam position is well-aligned\n"
			 "with the pixel-grid (rather than falling across part of each of 2 or 4 pixels). The true beam-center is given\n"
			 "(in X11-coordinates) by the Centroid position; this is rounded to the nearest integer to get the best pixel to\n"
			 "select for Hawaii-CDSM-readout. See 'cam2tiff -h' for a diagram.\n"
			 "\n"
			 "Note 2: The algorithm is as follows:\n"
			 " * Find the brightest pixel in the image. Let B = that brightness value.\n"
			 " * [Check whether it should be merged into an existing, adjacent star.]\n"
			 " * Iteratively \"grow\" a star around it: \n"
			 " *    - for each pixel adjacent to the star, if (brightness > B *  %4.2f), add it to the star.\n"
			 " *    - stop: when there are no adjacent pixels of sufficient brightness, or after %d pixels (considered a \"whiteout\").\n"
			 " * Remove this set of pixels from the pool.\n"
			 " * Repeat, till there are no non-zero pixels left, or until at most %d candidate beams have been found.\n"
			 " * For each \"star\", calculate the intensity-weighted position (centroid), then round and convert to Hawaii-coords.\n"
			 " * The \"Magic Numbers\" are #defined in the source: LUM_THRESHOLD=%4.2f, MAX_PIXELS_PERSTAR=%d, MAX_STARS=%d.\n"
			 " * There is a tradeoff in the value of LUM_THRESHOLD between 'collecting' all the star-material vs noise-immunity.\n"
			 "\n"
			 "Note 3: A side-benefit of the chosen grow-star algorithm is that it makes artefacts look obviously and visibly bad.\n"
			 "\n", MAX_STARS, MAX_PIXELS_PERSTAR, argv0, TIFF_EXTN, LUM_THRESHOLD, MAX_STARS, LUM_SCALEFACTOR_8BIT, LUM_THRESHOLD, MAX_PIXELS_PERSTAR, MAX_STARS, LUM_THRESHOLD, MAX_PIXELS_PERSTAR, MAX_STARS);
}


/* Open TIFF file, and read it into the pixels[][] array. */
void read_input_tiff (char *tiff_filename){
	unsigned int j;
	uint32 w, h, rps; uint16 bps, spp, orient;	/* tiff properties */
	TIFF *tiff;

	/* Open TIFF file */
	if( ( tiff = TIFFOpen ( tiff_filename, "r" ) ) == NULL ) {
		fprintf ( stderr, "Could not open %s for reading: %s\n", tiff_filename, strerror(errno) );
		exit ( 1 );
	}

 	/* Sanity-check the TIFF file - should match what cam2tiff creates! */
	TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &w);
	TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &h);
	TIFFGetField(tiff, TIFFTAG_BITSPERSAMPLE, &bps);
	TIFFGetField(tiff, TIFFTAG_SAMPLESPERPIXEL, &spp);
	TIFFGetField(tiff, TIFFTAG_ROWSPERSTRIP, &rps);
	TIFFGetField(tiff, TIFFTAG_ORIENTATION, &orient);

	if ( (w != WIDTH) || (h != HEIGHT) || (bps != TIFF_EXPECT_BPS) || (spp != TIFF_EXPECT_SPP) || (rps != TIFF_EXPECT_RPS) || (orient != TIFF_EXPECT_ORIENT) ){
		fprintf( stderr, "Malformed Tiff.\n"
				 "    Expect dimensions %d x %d, %d bits/sample, %d samples/pixel, %d rows/strip, orientation %d.\n"
				 "    Found: dimensions %d x %d, %d bits/sample, %d samples/pixel, %d rows/strip, orientation %d.\n",
				 WIDTH, HEIGHT, TIFF_EXPECT_BPS, TIFF_EXPECT_SPP, TIFF_EXPECT_RPS, TIFF_EXPECT_ORIENT,    w, h, bps, spp, rps, orient);
		exit (1);
	}

	/* Read successive strips (rows) into array. */
	for (j = 0 ; j < TIFFNumberOfStrips(tiff); j++) {
		TIFFReadEncodedStrip(tiff, j, pixels[j], sizeof(pixels[j]));
	}

	/* check it worked */
	// x = 1020; y = 10;
	// fprintf (stderr, "x = %u, y = %u, value = %u \n", x, y, pixels[y][x]);

	/* Close the file: */
	TIFFClose ( tiff );
}


/* Write out a normalised RGB output TIFF file */
void write_normalised_rgb_output_tiff (char *tiff_filename2){
	int i, j;
	float normaliser;
	TIFF *tiff2;

	/* Normalise factor */
	if (peak_for_normalising != 0){
		normaliser = floor( (float)UINT32_MAX / peak_for_normalising);
	}else{
		normaliser = 1.0;
	}

	fprintf (stderr, "\nPlotting the result. Normalised (by %4.2f) to 3x16-bit RGB; detected beams are overlaid in red.\n", normaliser);

	/* Create the output image. Iterate over the main one, normalise, convert to 16-bits, and mask the red-channel. */
	for (j=0; j < HEIGHT; j++){
		for (i=0; i < WIDTH; i++){
			output_b[j][i] = ((uint32)(normaliser * pixels[j][i])) >> (TIFF_EXPECT_BPS - TIFF_OUTPUT_BPS);
			//output_r[j][i] = output_g[j][i] = output_b[j][i];   /* Just show raw data in blue channel? Clearer? */
			if ( star_map [j][i] ){
				output_r[j][i] = 0xffff;   /* For detected "stars", set  red-channel to full, and zero the others. */
				output_g[j][i] = 0;	   /* Idea: zeroing just the green, but not the blue channel would allow  */
				output_b[j][i] = 0;   	   /*    more info to be conveyed, though perhaps losing clarity.  (TODO?)) */
			}				   /* Or maybe show the pixel data in the blue channel; mark strong detected stars as #f00, and use #0f0 for faint ones? */
		}
	}

	/* Open TIFF file for writing */
	if( ( tiff2 = TIFFOpen ( tiff_filename2, "w" ) ) == NULL ) {
		fprintf ( stderr, "Could not open %s for writing: %s\n", tiff_filename2, strerror(errno) );
		exit ( 1 );
	}


	/* Configure TIFF file */
	TIFFSetField ( tiff2, TIFFTAG_IMAGEWIDTH, WIDTH);			/* 1024 x 1024  (or num_cols*2 x num_rows*2) */
	TIFFSetField ( tiff2, TIFFTAG_IMAGELENGTH, HEIGHT);
	TIFFSetField ( tiff2, TIFFTAG_BITSPERSAMPLE, TIFF_OUTPUT_BPS );		/* 16-bit per sample, for each of R,G,B */
	TIFFSetField ( tiff2, TIFFTAG_SAMPLESPERPIXEL, TIFF_OUTPUT_SPP );	/* 3 samples per pixel (RGB) */
	TIFFSetField ( tiff2, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB );
	TIFFSetField ( tiff2, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB );		/* Note: the reverse, LSB2MSB is not widely supported; don't use it. */
	TIFFSetField ( tiff2, TIFFTAG_ORIENTATION, TIFF_EXPECT_ORIENT );	/* Many readers only support the value of TOPLEFT. Just as well that we can use this orientation. */
	TIFFSetField ( tiff2, TIFFTAG_ROWSPERSTRIP, TIFF_EXPECT_RPS );
	TIFFSetField ( tiff2, TIFFTAG_COMPRESSION, COMPRESSION_NONE );		/* Should we use COMPRESSION_DEFLATE? Probably not, if combining SSD with slow CPU */
	TIFFSetField ( tiff2, TIFFTAG_PLANARCONFIG, PLANARCONFIG_SEPARATE );	/* Important, for RGB. The image planes are R, G, B (not interleaved) */
	TIFFSetField ( tiff2, TIFFTAG_XRESOLUTION, 150.0 );			/* Arbitrary */
	TIFFSetField ( tiff2, TIFFTAG_YRESOLUTION, 150.0 );
	TIFFSetField ( tiff2, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH );

	/* Write successive strips (rows) from the array. */
	for (j = 0 ; j < HEIGHT; j++) {		/* Red channel */
		if ( TIFFWriteEncodedStrip(tiff2, j, output_r[j], sizeof(output_r[j])) < 0 ){
			fprintf ( stderr, "Error writing line %d to file %s: %s\n", j, tiff_filename2, strerror(errno) );
			exit (1);
		}
	}
	for (j = 0 ; j < HEIGHT; j++) {		/* Green channel. 2nd complete plane. */
		if ( TIFFWriteEncodedStrip(tiff2, (j + HEIGHT), output_g[j], sizeof(output_g[j]) ) < 0 ){
			fprintf ( stderr, "Error writing line %d to file %s: %s\n", j, tiff_filename2, strerror(errno) );
			exit (1);
		}
	}
	for (j = 0 ; j < HEIGHT; j++) {		/* Blue channel. 3rd complete plane. */
		if ( TIFFWriteEncodedStrip(tiff2, ( j + 2*HEIGHT), output_b[j], sizeof(output_b[j]) ) < 0 ){
			fprintf ( stderr, "Error writing line %d to file %s: %s\n", j, tiff_filename2, strerror(errno) );
			exit (1);
		}
	}

	/* Close tiff */
	TIFFClose ( tiff2 );
}

/* Starbit: return the bitmask for the nth star. */
unsigned int sbit(int bitnum){
	return (0x1 << bitnum);
}

/* Is a particular pixel address legal? i.e. Can we look up pixeks[y][x] without segfault or overflow? */
int is_legal (int y, int x){
	return (   (x >= 0) && (x < WIDTH)   &&  ( y >= 0) && (y < HEIGHT)   );
}

/* Find adjacent star to a pixel, if there is one. Return the star_number (star_n) if so, -1 if not, -2 if too big already. (If there are multiple stars, return the lowest-numbered one). */
int find_adjacent_star (int j, int i){
	unsigned int star_bits = 0;
	int x,y, star_n = -1;

	x = i-1 ; y = j-1;		/* Check 8 neighbours, do bitwise or. */
	if (is_legal(y, x) ){
		star_bits |=  star_map[y][x]; }

	x = i-1 ; y = j;
	if (is_legal(y, x) ){
		star_bits |=  star_map[y][x]; }

	x = i-1 ; y = j+1;
	if (is_legal(y, x) ){
		star_bits |=  star_map[y][x]; }

	x = i ; y = j-1;
	if (is_legal(y, x) ){
		star_bits |=  star_map[y][x]; }

	x = i ; y = j+1;
	if (is_legal(y, x) ){
		star_bits |=  star_map[y][x]; }

	x = i+1 ; y = j-1;
	if (is_legal(y, x) ){
		star_bits |=  star_map[y][x]; }

	x = i+1 ; y = j;
	if (is_legal(y, x) ){
		star_bits |=  star_map[y][x]; }

	x = i+1 ; y = j+1;
	if (is_legal(y, x) ){
		star_bits |=  star_map[y][x]; }

	for (star_n = 0; star_n < MAX_STARS; star_n++){
		if ( (star_bits >> star_n) & 0x1){	/* This star is ajacent to the point */
			if (stars[star_n].pixels == MAX_PIXELS_PERSTAR){
				return (-2);		/* But that star is already at its limit (otherwise, we get very big stars, and, a really slow algorithm). */
			}else{
				return (star_n);	/* So...append to this star */
			}
		}
	}
	return (-1);					/* No star is adjacent to point (y=y,x=i) */
}


/* Find the brightest as-yet-unvisited pixel. If found, place it into the star_map[][] as nth-bit-bitmask. If not found, return 0 to break out of the loop *
 * Naive algorithm: not ideal in the case when a star falls between/across 2-4 overlapping px, though it will work OK for us, unless the S/N ratio is really awful */
/* When find_brightest_star locates a new pixel, it should test whether it is adjacent to any of the pixels in any of the current stars. If so, merge with them. */
int find_brightest_star(){
	int i,j,x = 0, y = 0, star_n;
	uint32 brightest = 0;

	for (j=0; j < HEIGHT; j++){	/* Iterate over rows and cols... */
		for (i=0; i < WIDTH; i++){
			if (( pixels[j][i] > brightest) && (! star_map[j][i]) ){  /* Is it the brightest so far, and not already taken? */
				brightest = pixels[j][i];
				x = i; y = j;
			}
		}
	}
	if (brightest > 0){				/* Did we find one? */
							/* YES */
		star_n = find_adjacent_star(y, x);	/* 	Check in case it already touchses an existing star (the "split-stars problem") */

		if (star_n >= 0){			/*     		Yes, this touches the existing star_n. Add it to that star, return(-2) so grow_star is not triggered */
			star_map [y][x] = sbit(star_n);
			debug_printf ("found brightest pixel at x=%d, y=%d, adjacent to existing star, so merged with star n=%d.\n", x,y,star_n);
			return (-2);

		}else{	/* -1 or -2 */			/*		No, this is a new star. Put it into the next bitfield fof starmap[][], and increment the counter. */
			star_n = stars_found;		/*		(or, if -2, we can't grow into an existing star which is at limiting size, so must treat this as new. */
			stars_found++;
			star_map [y][x] = sbit(star_n);
			if (peak_for_normalising == 0){	/*  		[First time only: save the brightest overall value, for later normalisation.] */
				peak_for_normalising = brightest;
			}
			debug_printf ("found star n = %d at x = %d, y = %d, brightness %u\n", star_n, x, y, brightest);
			return (star_n);
		}
	}else{
		return (-1);				/* NO.   Terminate the search. */
	}
}


/* Checkpixel: helper function for grow_star(). Checks whether a particular point should be accreted into the star-mass */
int checkpixel (int x, int y, uint32 threshold, int pixelcount, int star_n __attribute__ ((unused)) ){
	debug_printf ("x is %d, y is %d, thresh is %u, pixcount is %d,  value is %lld\n", x,y, threshold, pixelcount, (((x>= 0)&&(x<WIDTH)&&(y>= 0)&&(y<HEIGHT))?(long long)pixels[y][x]:-1) ); //ternary thing: prevent segfault when debugging!

	if( (is_legal(y,x) ) &&				/* Check it's legal: else we are outside the image, and (at best) about to segfault! */
	    (! star_map[y][x] ) &&			/* Check the pixel isn't already in the current star's set, nor that of any other star (i.e. that bitfield == 0.) */
	    (pixels[y][x] >= threshold) ){		/* Is the pixel at least 3% as bright as the one that is is the central point in the star? */
							/* NB the bounds of x and y must be checked FIRST (using short-circuiting '&&'), else the lookup in current[][] could segfault */
		debug_printf ("added this pixel to star %d\n", star_n );
		if (pixelcount +1 > MAX_PIXELS_PERSTAR){
			debug_printf ("Error: grow_star has now reached %d pixels; this exceeds the configured maximum of %d pixels. Suspected whiteout! \n", pixelcount, MAX_PIXELS_PERSTAR);
			return (0);
		}else{
			return (1);
		}
	}
	return (0);
}

/* "Grow" the star: add possible star-matter by searching adjacent points to see if they are a decent-fraction of the current brightness */
/* Note that this is an inefficient (but still fast if MAX_PIXELS_PERSTAR is small) algorithm: it repeatedly scans the same pixels, even those in the middle of the beam (it's O2 !) */
/* Consider looking at background noise level? Consider looking for areas of high correlation? */
/* Consider tapering the threshold wrt radius from center - but this would make linear artefacts look more like stars - perhaps undesirable */
/* Also note that the current method always seeks to the right first. If it finds a bogus star with high surrounding noise, it will tend keep drawing towards the right, */
/* than drawing a circle. This is accidentally good, because it makes artefacts look bad */
/* NB if lum_threshold is too high, we tend to collect up "noise", and go towards a whiteout. */
/*    if too low, we fail to gather all of a star, but leave a relatively bright pixel behind, which may seed a subsequent star touching this one. */
/* NB This assumes that it starts with a single-pixel star, and that single-pixel is the brightest one. */
void grow_star(int star_n){
	int x,y,i,j, pixelcount_prev;
	int pixelcount = 1;
	unsigned int bit;
	uint32 brightness = 0, threshold = 0;

	bit = sbit(star_n);	/* Bitmask in star_map[][] */

	while (1){	/* Keep looping...  */
		pixelcount_prev = pixelcount;
		for (j=0; j < HEIGHT; j++){	/* Iterate over rows and cols of the current star's star_map ... */
			for (i=0; i < WIDTH; i++){
				if (star_map[j][i] == bit){
					if (brightness == 0){   /* Find the brightness threshold as a given fraction of the brightness of the first (and brightest) pixel */
						brightness = pixels[j][i];    /* NB, don't set threshold wrt the current edge of the star, else we may grow forever */
						threshold = brightness * lum_threshold;
					}

					/* For each of the 8 adjacent pixels, check whether we can "grow" the star to include it. */
					x = i-1 ; y = j-1;
					if (checkpixel(x, y, threshold, pixelcount, star_n)){
						star_map[y][x] = bit;   pixelcount++; }     /* yes...add it to star_map[][]'s bitfield  */

					x = i-1 ; y = j;
					if (checkpixel(x, y, threshold, pixelcount, star_n)){
						star_map[y][x] = bit;   pixelcount++; }

					x = i-1; y = j+1;
					if (checkpixel(x, y, threshold, pixelcount, star_n)){
						star_map[y][x] = bit;   pixelcount++; }

					x = i; y = j-1;
					if (checkpixel(x, y, threshold, pixelcount, star_n)){
						star_map[y][x] = bit;   pixelcount++; }

					x = i; y = j+1;
					if (checkpixel(x, y, threshold, pixelcount, star_n)){
						star_map[y][x] = bit;   pixelcount++; }

					x = i+1; y = j-1;
					if (checkpixel(x, y, threshold, pixelcount, star_n)){
						star_map[y][x] = bit;   pixelcount++; }

					x = i+1; y = j;
					if (checkpixel(x, y, threshold, pixelcount, star_n)){
						star_map[y][x] = bit;   pixelcount++; }

					x = i+1; y = j+1;
					if (checkpixel(x, y, threshold, pixelcount, star_n)){
						star_map[y][x] = bit;   pixelcount++; }
				}
			}
		}
		if (pixelcount == pixelcount_prev){	/* ...this iteration added no more star matter. Stop repeating now. */
			break;
		}
	}

	stars[star_n].n = star_n;			/* Begin to populate star-struct. */
	stars[star_n].pixels = pixelcount;
}


/* Calculate the statistics for the nth star. */
void star_statistics(int star_n){
	int i,j, n, x, y, h_q, h_x, h_y, i_maxbri = 0, j_maxbri = 0;
	unsigned int bit;
	uint32 brightness = 0;
	int pixelcount = 0;
	long long luminosity = 0;
	double c_x, c_y, focus, sigma_l_x = 0, sigma_l_y = 0;
	char *valid;

	n = star_n;
	bit = sbit (n);

	for (j=0; j < HEIGHT; j++){	/* Iterate over rows and cols of this star ... */
		for (i=0; i < WIDTH; i++){
			if (star_map[j][i] == bit){
				pixelcount++;			/* add up star pixels and luminosity. */
				luminosity += pixels[j][i];

				if (pixels[j][i] > brightness){		/* Get coordinates of the brightest point in this star */
					brightness = pixels[j][i];  	/* [probably the first pixel we found; might be an adjacent point if 2 pixels are exactly equal] */
					i_maxbri = i; j_maxbri = j;
				}

				sigma_l_x += (double)pixels[j][i] * i;
				sigma_l_y += (double)pixels[j][i] * j;
			}
		}
	}

	c_x = sigma_l_x / luminosity;	/* centroid: "center of mass" */
	c_y = sigma_l_y / luminosity;

	x = round (c_x);		/* x,y are the integer-pixel positions: the best single pixel points for Hawaii alignment */
	y = round (c_y);

  	/* Central focus: the fraction of the total luminosity that falls into the INTEGER pixel identified as "the" center */
	/* This is a measure of how well the optics are aligned with the hawaii-grid, and how well the beam is targeted to one pixel */
	/* [It's not really intended to measure how "peaky" the light distribution is (that would need to look at the centroid */
	focus = (double) pixels[y][x] / luminosity;

	if ((x < WIDTH/2) && (y < WIDTH/2)){	/* Convert (integer) X11 coordinates to Hawaii coordinates. See `cam2tiff -h` for a diagram. */
		h_q = 2;
		h_x = x;
		h_y = (HEIGHT/2) - 1 - y;
	}else if ((x < WIDTH/2) && (y >= HEIGHT/2)){
		h_q = 1;
		h_x = x;
		h_y = (HEIGHT) - 1 - y;
	}else if ((x >= WIDTH/2) && (y < HEIGHT/2)){
		h_q = 3;
		h_x = x - (WIDTH/2);
		h_y = (HEIGHT/2) - 1 - y;
	}else{
		h_q = 0;
		h_x = x - (WIDTH/2);
		h_y = (HEIGHT) - 1 - y;
	}

	/* Is the beam probably valid? Check whether there are too many pixels (> PIXELCOUNT_VALID), whether we hit a whiteout (> MAX_PIXELS_PERSTAR), */
	/* whether we have enough light in this beam at all (> VALID_MINRELPXBRI * brightestpoint), and whether the focus is good enough (> VALID_FOCUSMIN) */
	if ( (pixelcount < MAX_PIXELS_PERSTAR) &&  (pixelcount <= VALID_MAXPIXELCOUNT) &&
	     (brightness >= (VALID_MINRELPXBRI * peak_for_normalising) ) && (focus >= VALID_FOCUSMIN) ){
		valid = "yes";
	}else{
		valid = "no";
	}

 	stars[n].n = n;				/* Save struct into array, for subsequent printout. */
 	stars[n].x = i_maxbri;
	stars[n].y = j_maxbri;
	stars[n].brightest = brightness;
	stars[n].c_x = c_x;
	stars[n].c_y = c_y;
	stars[n].h_q = h_q;
	stars[n].h_x = h_x;
	stars[n].h_y = h_y;
	stars[n].pixels = pixelcount;
	stars[n].total_lum = luminosity;
	stars[n].focus = focus;
	stars[n].avg_lum = luminosity / pixelcount;
	stars[n].valid = valid;
}


/* Start here... */
int main ( int argc, char **argv ) {
	int  opt; extern char *optarg; extern int optind, opterr, optopt;  /* getopt */
	char *tiff_filename;
	char cmd[128] = "" ;
	char tiff_filename2[] = "/tmp/beamfinder_masked.tiff.XXXXXX"; /* For mkstemp() */
	int n, ret, star_n = 0, do_plot = 0;

	/* Sanity check #defines */
	if (MAX_STARS > 8 * sizeof (int)){
		printf ("Config Error: MAX_STARS is #define'd as %d; this is too big to fit an integer bitmask.\n", MAX_STARS);
		exit(1);
	}else if (MAX_STARS * MAX_PIXELS_PERSTAR > WIDTH * HEIGHT){
		printf ("Config Error: MAX_STARS * MAX_PIXELS_PERSTAR > WIDTH * HEIGHT  (#define'd as %d, %d, %d, %d); algorithm not guaranteed to terminate.\n", MAX_STARS, MAX_PIXELS_PERSTAR, WIDTH, HEIGHT);
		exit(1);
	}

	/* Get options and check for validity */
	if ((argc > 1) && (!strcmp (argv[1], "--help"))) {	/* Support --help, without the full getopt_long */
		print_help(argv[0]);
		exit (EXIT_SUCCESS);
	}

	while ((opt = getopt(argc, argv, "hpt:")) != -1) {	/* Getopt */
		switch (opt) {
			case 'h':				/* help */
				print_help(argv[0]);
				exit(EXIT_SUCCESS);
				break;
			case 'p':				/* plot: show image and detected beams. */
				do_plot = 1;
				break;
			case 't':				/* change luminosity threshold from the default */
				lum_threshold = atof(optarg);
				break;
			default:				/* Catches wrong options or options missing their arg */
				fprintf ( stderr, "Syntax: %s [OPTIONS] input.%s     #use -h for help\n", argv[0], TIFF_EXTN );
				exit (EXIT_FAILURE);
		}
	}

	if (argc - optind != 1){
		fprintf ( stderr, "Syntax: %s [OPTIONS] input.%s    #use -h for help\n", argv[0], TIFF_EXTN );
		exit ( 1 );
	}

	/* Check tiff_filename ends in .tiff (or tif). Useful to prevent overwriting the wrong file by accident.  */
	tiff_filename = argv[optind];
	if ( (strcmp ( (tiff_filename + strlen(tiff_filename) - strlen(TIFF_EXTN) -1), "."TIFF_EXTN) )  &&  (strcmp ( (tiff_filename + strlen(tiff_filename) - strlen(TIF_EXTN) -1), "."TIF_EXTN) )  ){
		fprintf(stderr,"Error, input file %s is not a .%s (or .%s) file. (Wrong extension).\n", tiff_filename, TIFF_EXTN, TIF_EXTN);
		exit (1);
	}


	/* Open TIFF file, and read it into the pixels[][] array. */
	read_input_tiff (tiff_filename);


	/* Iterate over the image. Stop after finding MAX_STARS, or if there are no more to find. */
	while ( stars_found < MAX_STARS ){

		star_n = find_brightest_star(); 	/* Find the brightest as-yet-unvisited pixel. Returns star_n, i.e. the index of the nth star, or -1 if not found. */
		if (star_n == -1){			/*  ...end the loop if no more to find */
			break;
		}else if (star_n == -2){		/*  ...pixel has been merged into an existing star; skip  grow_star() */
			continue;
		}

		grow_star(star_n);			/* "Grow" the star: add possible star-matter by searching in torus. */
	}

	/* Process stars: calculate the statistics for them */
	for (star_n = 0; star_n < stars_found; star_n++){
		star_statistics (star_n);
	}

	/* Print out the (formatted) results, stored in the stars[] array */
	printf ("#Found %d candidate beams, in descending order of luminosity:\n", stars_found);
	printf ("#Hawaii_Coordinates        Beam       Focus &       Valid     Beam Total       Average          X11_Coordinates (topleft)\n");
	printf ("#QUAD   X       Y          Pixels     Alignment     Beam?     Luminosity       Luminosity       Centroid Position (X,Y)\n");

	for (n = 0 ; n < stars_found; n++){
		printf ("%d      %3d     %3d         %3d        %4.2f          %-3s     %12lld    %14.2f       %8.2f     %8.2f\n",
			stars[n].h_q, stars[n].h_x, stars[n].h_y, stars[n].pixels, stars[n].focus, stars[n].valid, (stars[n].total_lum/LUM_SCALEFACTOR_8BIT),
			(stars[n].avg_lum/LUM_SCALEFACTOR_8BIT), stars[n].c_x, stars[n].c_y);  /* Scale luminosity to Hawaii ADC units */
	}

	/* Plot the result: show the original, with the detected areas in red */
	if (do_plot == 1){

		/* Create tempfile for output */
		if (mkstemp(tiff_filename2) == -1){	/* Create tempfile from template. NB TIFFOpen wants the fileNAME not a fileHANDLE */
			fprintf(stderr, "mkstemp failed with template %s: %s\n", tiff_filename2, strerror(errno) );
			exit(1);
		}

		/* Write out a normalised RGB output TIFF file */
		write_normalised_rgb_output_tiff (tiff_filename2);

		/* Show it, using "display" */
		fprintf(stderr, "(Using 'display' to show tempfile %s , which will then be deleted.)\n", tiff_filename2);
		sprintf (cmd, "%s -title \"Found %d beams\" %s", DISPLAY_PROGRAM, stars_found, tiff_filename2);   /* Build the command (predefined string, of known length) */
		ret = system (cmd);
		if (ret != 0){
			fprintf (stderr, "Error occurred running command. Retval: %d, Cmd was: '%s'\n", ret, cmd);
		}

		/* Delete the file */
		unlink (tiff_filename2);
	}

	/* if we get here, we succeeded. */
	fprintf(stderr, "\n");
	return (0);
}
