/* Convert camera data format into a tiff. Invoke with -h for help and explanation */

/* Copyright (C) Richard Neill 2006-2011, <cam2tiff at REMOVE.ME.richardneill.org>. This program is Free Software. You can
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
 * For the details, please see: http://www.gnu.org/licenses/gpl.html */

/* Note: this uses native Intel format for the .cam format (32-bit little-endian, with MSByte zeroed), and then outputs a standard (big-endian) tiff. */
/* Byteswapping is done automatically by libtiff as necessary, depending on TIFFTAG_FILLORDER */

	/* Num rows/cols has a hardcoded maximum, but use -r and -c to deal with subarrays (i.e. not reading out the entire chip). */

/* Tools:  'Gimp' can't handle 32-bit tiffs. 'Convert' can convert them to .jpg; 'display' can display the image (and has some extra features for viewing) */

/* Statistics: we really want to use 128-bit ints, but this isn't possible: '__uint128_t' can only be used on 64-bit systems. Can use libgmp, but that's just excessively complex */
/* Instead, do it "by hand". Use uint64_t as our highest precision datatype (8 byte, unsigned-long-long), and manually handle wrapping */
/* (It's OK to use floats (long double) for the final calculations, but not for the summation, as that causes large cumulative errors.) */
/* WARNING: this will go wrong if our raw datatype ever exceeds 24-bits ! */
/* Raw data:    24-bit									*/
/* Num_samples: 20 bit => uint32				[each quad is 18 bit]	*/
/* Sum:		24 bit x 20 bit  =  44 bit => uint64  		[each quad is 42 bit] 	*/
/* Square:      48 bit => uint64							*/
/* Sum_squares: 48 bit X 20 bit = 68 bit =>   int:uint64	[each quad is 68 bit]	*/
/* Mean, Variance: use floating point: long double (12 bytes)				*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <tiffio.h>
#include <math.h>

#define eprintf(...)	fprintf(stderr, __VA_ARGS__)				/* Error printf: send to stderr  */
#define feprintf(...)	fprintf(stderr, __VA_ARGS__); exit (EXIT_FAILURE)	/* Fatal error printf: send to stderr and exit */

#define  DEBUG	   0

#define  CAM_EXTN  "cam"		/* Extensions: cam/dat for infile; tif/tiff for outfile */
#define  DAT_EXTN  "dat"
#define  TIF_EXTN  "tif"
#define  TIFF_EXTN "tiff"

#define NUM_ROWS 512			/* Number of rows, per quadrant (maximum possible) */
#define NUM_COLS 512			/* Number of columns, per quadrant (maximum possible) */
#define QUADRANTS 4			/* There are 4 quadrants */
#define HALVES 2			/* There are 2 row-halves in a row */

typedef uint32_t sample_t;		/* 4 bytes per sample */

unsigned int num_rows = NUM_ROWS;	/* Actual number of rows, may be overridden by -r */
unsigned int num_cols = NUM_COLS;	/* Actual number of cols, may be overridden by -c */

int errmsg_once = 0;			/* show errormsg only once */
struct stat stat_p;         		/* pointer to stat for filesize. */

/* Statistics: n */
uint32_t	num_samples, num_samples_q[QUADRANTS];
uint64_t	sum, sum_q[QUADRANTS];
uint64_t	square;
uint64_t	sum_squares, sum_squares_q[QUADRANTS];
int		sum_squares_hi, sum_squares_hi_q[QUADRANTS]; /* high bytes */
long double	sum_squares_f, sum_squares_f_q[QUADRANTS], mean, mean_q[QUADRANTS], variance, variance_q[QUADRANTS];

void print_help() {
	printf ("cam2tiff converts the data file from the IR camera into a tiff file.\n\n"
		"USAGE:    cam2tiff [OPTIONS] input.%s output.%s           #Not stdin/stdout.\n"
		"OPTIONS:     -h, --help        show help\n"
		"             -c,-W    COLS     number of columns (width) for each quadrant. Default: %d\n"
		"             -r,-H    ROWS     number of rows (height) for each quadrant. Default: %d\n\n"
		"The camera contains %dx%d pixels, arranged in %d quadrants of %dx%d pixels.\n"
		"These quadrants are Q0,Q1,Q2,Q3 (alternatively labelled 1,2,3,4) positioned as\n"
		"Q0=bottom_right. Q1=bottom_left, Q2=top_left, Q3=top_right.\n"
		"\n"
		"The input.cam file is a 4 MByte file (%d quadrants, each %dx%d pixels; %d-Byte words).\n"
		"The data is an interleaved stream of pixels from each quadrant in turn (0,1,2,3).\n"
		"A pixel is a 24-bit ADC result (~ 19:5 bits signal:noise), in the 3 LSB of a little-endian uint32.\n"
		"Thus the byte order is: Q0_L:M:H:0, Q1_L:M:H:0; Q2_L:M:H:0; Q3_L:M:H:0.\n"
		"\n"
		"These %d words form an interleaved raster, scanning left->right (fast) and bottom->top (slowly).\n"
		"cam2tiff re-arranges this muddle into a single %dx%d pixel, 32-bit greyscale TIFF file,\n"
		"with Q0 in the bottom right. For displayability, bytes are left-shifted: LSByte is zero.\n"
		"Here is a diagram; see the file 'llcpad.html' in the Hawaii documentation for a better one:\n"
		"\n"
		"  (y)     +---------------------+        \n"
		"          |          |         z|        Pixels are read in the order\n"
		"   s ^    |          |          |        abcdefgh...pqrs...z\n"
		"   l |    |    Q2    |    Q3    |        \n"
		"   o |    |r         |s         |        Slow scan, fast scan refer to the\n"
		"   w |    |cg        |dh        |        clocking of the HAWAII chip.   \n"
		"     |    |----------+----------|        The fast scan (column / x) direction \n"
		"   s |    |          |          |        is controlled by PIXEL_CLOCK;  \n"
		"   c |    |          |          |        the slow scan (row / y) direction  \n"
		"   a |    |    Q1    |    Q0    |        is controlled by LINE_CLOCK.  \n"
		"   n |    |q         |p         |            \n"
		"          |bf        |ae        |        The output TIFF has the same\n"
		"          O---------------------+        orientation as the Hawaii chip.\n"
		"                                         But note that X11 considers (x,y) as\n"
		"                --------->               the directions (right,down).\n"
		"                 fast scan  (x)            \n"
		"\n"
		"\n"
		"The orientation of the .cam file (and the TIFF on screen) have the origin (O) in the bottom left, \n"
		"(i.e. x-axis (col) points to the right, and y-axis (row) points upward; Q0 is in bottom right.)\n"
		"It is displayed on screen this way too. However, the X11 convention is that the origin is top-left,\n"
		"the x-axis points right, and y-axis points down. [Both conventions are 0-based. So, pixel co-ordinates\n"
		"as show by 'display', or used to specify geometry for 'stream' differ from Hawaii co-ordinates.\n"
		"To transform from one co-ordinate system to the other, use: x' = x ; y' = (%d - y) .\n"
		"If only a sub-array of the chip is being read out, use -r, -c to override the default array-size.\n"
		"\n"
		"Statistics: mean, std-dev (population), number. These are calculated for each quadrant, and overall.\n"
		"Exit codes: 0 on success; 1 on failure; 2 on error from which some data was recovered.\n"
		"Viewing/processing 32-bit tiffs: use 'display'/'stream'; see also tiffcp. Currently, all of gimp, eog, \n"
		"qiv fail on 32-bit images. NB, build ImageMagick --with-quantum-depth=32.\n"
		"\n",
		CAM_EXTN, TIFF_EXTN, NUM_ROWS, NUM_COLS, NUM_COLS*2, NUM_ROWS*2, QUADRANTS, NUM_COLS, NUM_ROWS, QUADRANTS, NUM_COLS,
		NUM_ROWS, (int)sizeof(sample_t), QUADRANTS, NUM_COLS*2, NUM_ROWS*2, NUM_COLS*2-1
		);
}


static void process_row ( FILE *cam, TIFF *tiff, unsigned int row ) {
	sample_t sample;
	sample_t chunk_buf[NUM_COLS * QUADRANTS];	/* A chunk is one row, but containing 4 quadrant-interleaved samples. */
	sample_t top_row[NUM_COLS * HALVES];
	sample_t bottom_row[NUM_COLS * HALVES];
	long file_offset;
	unsigned int col, quad;
	unsigned int chunksize;

	chunksize = num_cols * QUADRANTS * sizeof(sample_t);  /* Usually sizeof(chunk_buf), but smaller if -c specifies other than the maximum, */
	file_offset = row * chunksize;

	if ( fseek ( cam, file_offset, SEEK_SET ) != 0 ) {
		feprintf ( "Could not seek to offset %ld: %s\n", file_offset, strerror ( errno ) );
	}

	if ( fread ( chunk_buf, 1, chunksize, cam ) != chunksize ) { /* Read in data in intel-native format */
		if ( feof ( cam ) ) {
			if (errmsg_once == 0){
				errmsg_once=1;  /* I shall say this only once */
				eprintf ( "Premature EOF on input. Expected input file of exactly %dx%d pixels x %d bits, i.e. %.2f MBytes.\n", num_cols*2, num_rows*2, (int)sizeof(sample_t)*8, ((double)num_cols*2 * num_rows*2 * sizeof(sample_t) / (1024*1024) ) );
				eprintf	( "However, the input file is only %ld bytes, (i.e. %ld pixels) long; failed on row %d.\n", stat_p.st_size, stat_p.st_size/ sizeof(sample_t), row);
				eprintf	( "Generating a CORRUPT tiff file, which might still be slightly useful (try opening it in the GIMP). Or specify the array size with -r and -c. \n");
			}
			return;		/* might be useful to generate a corrupt tiff, rather than giving up. So return, rather than exit. Also, if we exit here, the previous tiff will remain. */
		} else {
			feprintf ( "Error reading: %s\n", strerror ( errno ) );  /* always a fatal error */
		}
	}

	/* De-multiplex the stream into 4 substreams. Left-shift. Do statistical analysis */
	for ( col = 0 ; col < num_cols ; col++ ) {
 		bottom_row[num_cols + col] = chunk_buf [(col*4) + 0] <<8;   /* Demultiplex: Bottom_row = Q1:Q0    Top_row = Q2:Q3 */
 		bottom_row[col] 	   = chunk_buf [(col*4) + 1] <<8;   /* left-shift, so that the TIFF can be sensibly displayed, rather than having a zero MSByte */
 		top_row   [col]            = chunk_buf [(col*4) + 2] <<8;
 		top_row   [num_cols + col] = chunk_buf [(col*4) + 3] <<8;

		for (quad = 0; quad < QUADRANTS; quad++){
			sample = chunk_buf [(col*4) + quad];

			num_samples_q[quad] ++;				/* Now do stats (with the unshifted data) */
 			sum_q[quad] += sample;
 			square = (uint64_t) sample * (uint64_t) sample;

 			sum_squares_q[quad] += square;			/* Add up, and manually carry */
 			if (sum_squares_q[quad] < square){
 				sum_squares_hi_q[quad] ++;
 			}
		}
	}

	/* Now process these demultiplexed strips into a TIFF. */
	if ( TIFFWriteEncodedStrip ( tiff, ( ( 2 * num_rows - 1 ) - row ), &bottom_row, (num_cols * 2 * sizeof (sample_t)) ) < 0 ) {
		feprintf ( "Error writing: %s\n", strerror ( errno ) );
	}
	if ( TIFFWriteEncodedStrip ( tiff, ( ( num_rows - 1 ) - row ), &top_row, (num_cols * 2 * sizeof (sample_t)) ) < 0 ) {
		feprintf ( "Error writing: %s\n", strerror ( errno ) );
	}
}

int main ( int argc, char **argv ) {
	int  opt; extern char *optarg; extern int optind, opterr, optopt;  			/* getopt */
	char *cam_filename;
	char *tiff_filename;
	FILE *cam;
	TIFF *tiff;
	unsigned int row;
	unsigned int i;

	/* Get options and check for validity */
	if ((argc > 1) && (!strcmp (argv[1], "--help"))) {	/* Support --help, without the full getopt_long */
		print_help();
		exit (EXIT_SUCCESS);
	}

	while ((opt = getopt(argc, argv, "hc:W:r:H:")) != -1) {	/* Getopt */
		switch (opt) {
			case 'c':
			case 'W':
				num_cols  = atoi(optarg);	/* number of columns */
				break;
			case 'h':				/* help */
				print_help();
				exit(EXIT_SUCCESS);
				break;
			case 'r':
			case 'H':
				num_rows  = atoi(optarg);	/* number of rows */
				break;
			default:				/* Catches wrong options or options missing their arg */
				feprintf ( "Syntax: %s [OPTIONS] input.%s output.%s.   #use -h for help\n", argv[0], CAM_EXTN, TIFF_EXTN );
		}
	}

	if (argc - optind != 2){
		feprintf ( "Syntax: %s [OPTIONS] input.%s output.%s.   #use -h for help\n", argv[0], CAM_EXTN, TIFF_EXTN );
	}

	/* Files. NB can't use stdin here because of fseek(), (also would need to skip the filesize check). Can't use stdout because libtiff seeks. */
	cam_filename = argv[optind];
	tiff_filename = argv[optind + 1];

	/* Check num_cols and num_rows are sane */
	if ( (num_cols < 1) || (num_cols > NUM_COLS) ){
		feprintf( "Error, number of cols ('-c %d') must be between 1 and %d.\n", num_cols, NUM_COLS);
	}else if ( (num_rows < 1) || (num_rows > NUM_ROWS) ){
		feprintf( "Error, number of rows ('-r %d') must be between 1 and %d;\n", num_rows, NUM_ROWS);
	}

	/* Check cam_filename ends in .cam (or dat). Useful to prevent overwriting the wrong file by accident. */
	if (  (strcmp ( (cam_filename + strlen(cam_filename) - strlen(CAM_EXTN) -1), "."CAM_EXTN) )  &&  (strcmp ( (cam_filename + strlen(cam_filename) - strlen(DAT_EXTN) -1), "."DAT_EXTN) )  ){
		feprintf( "Error, input file %s is not a .%s (or .%s file). (Wrong extension).\n", cam_filename, CAM_EXTN, DAT_EXTN);
	}

	/* Open camera file. Can't use stdin here: we need to stat() the file to obtain its size and know the number of pixels. */
	if ( ( cam = fopen ( cam_filename, "r" ) ) == NULL ) {
		feprintf ( "Could not open %s for reading: %s\n", cam_filename, strerror(errno) );
	}

	if (stat (cam_filename, &stat_p) == -1){  					/* Get stats for file */
		feprintf( "Error occurred attempting to stat %s\n", cam_filename);	  /* This error shouldn't happen, since we already managed to open the file! */
	}

	if (stat_p.st_size == 0){							/* Check it isn't empty. */
		feprintf( "Error, input file %s is empty.\n", cam_filename);
	}

	/* Tiff file. Check tiff_filename ends in .tiff (or tif). NB can't use stdout here: libtiff requires random access. */
	if ( (strcmp ( (tiff_filename + strlen(tiff_filename) - strlen(TIFF_EXTN) -1), "."TIFF_EXTN) )  &&  (strcmp ( (tiff_filename + strlen(tiff_filename) - strlen(TIF_EXTN) -1), "."TIF_EXTN) )  ){
		feprintf( "Error, output file %s is not a .%s (or .%s) file. (Wrong extension).\n", tiff_filename, TIFF_EXTN, TIF_EXTN);
	}

	/* Open TIFF file */
	if( ( tiff = TIFFOpen ( tiff_filename, "w" ) ) == NULL ) {
		feprintf ( "Could not open %s for writing: %s\n", tiff_filename, strerror(errno) );
	}

	/* Configure TIFF file */
	TIFFSetField ( tiff, TIFFTAG_IMAGEWIDTH, num_cols * 2 );			/* 1024 x 1024  (or num_cols*2 x num_rows*2) */
	TIFFSetField ( tiff, TIFFTAG_IMAGELENGTH, num_rows * 2 );
	TIFFSetField ( tiff, TIFFTAG_BITSPERSAMPLE, 8 * sizeof ( sample_t ) );		/* 32-bit */
	TIFFSetField ( tiff, TIFFTAG_SAMPLESPERPIXEL, 1 );				/* Greyscale */
	TIFFSetField ( tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK );
	TIFFSetField ( tiff, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB );			/* Note: the reverse, LSB2MSB is not widely supported; don't use it. */
	TIFFSetField ( tiff, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT );		/* Many readers only support the value of TOPLEFT. Just as well that we can use this orientation. */
	TIFFSetField ( tiff, TIFFTAG_ROWSPERSTRIP, 1 );
	TIFFSetField ( tiff, TIFFTAG_COMPRESSION, COMPRESSION_NONE );
	TIFFSetField ( tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG );
	TIFFSetField ( tiff, TIFFTAG_XRESOLUTION, 150.0 );				/* Arbitrary */
	TIFFSetField ( tiff, TIFFTAG_YRESOLUTION, 150.0 );
	TIFFSetField ( tiff, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH );

	/* Process the camera data in (usually) 512-sample, 4-quad interleaved, blocks */
	for ( row = 0 ; row < num_rows ; row++ ) {
		process_row ( cam, tiff, row );
	}

	TIFFClose ( tiff );
	fclose ( cam );

	if ( abs(stat_p.st_size) > num_cols*2 * num_rows*2 * sizeof(sample_t)){
		eprintf ( "WARNING: discarded surplus data on input. Expected input file of exactly  %dx%d pixels x %d bits, i.e. %.2f MBytes.\n", num_cols*2, num_rows*2, (int)sizeof(sample_t)*8, ((double)num_cols*2 * num_rows*2 * sizeof(sample_t) / (1024*1024) ) );
		eprintf	( "However, the input file is actually %ld bytes, (i.e. %ld pixels) long; %ld bytes were ignored at the end.\n", stat_p.st_size, stat_p.st_size/sizeof(sample_t), stat_p.st_size - (num_cols*2 * num_rows*2 *sizeof(sample_t)) );
		errmsg_once=1;  /* Make sure we don't exit (0) */
	}

	/* Statistics. Get totals, then finish calculation of mean and (population) variance */
	for (i=0; i < QUADRANTS; i++){
		if (num_samples_q[i] == 0){
			feprintf ("Error: quadrant %d has zero samples. Can't do statistics.\n", i);
		}
		num_samples += num_samples_q[i];
		sum += sum_q[i];

		sum_squares += sum_squares_q[i];		/* Add up, and manually carry */
		if (sum_squares < sum_squares_q[i]){
			sum_squares_hi ++;
		}
		sum_squares_hi += sum_squares_hi_q[i];

		/* Floats: ok here, as errors aren't cumulative */
		sum_squares_f_q[i] = (long double) sum_squares_hi_q[i] *  pow (2,(sizeof(sum_squares_q[0])*8))  + sum_squares_q[i];
		mean_q[i] = sum_q[i] / num_samples_q[i];
		variance_q[i] = (sum_squares_f_q[i] - (((long double)sum_q[i] * sum_q[i]) / num_samples_q[i])) / num_samples_q[i];
	}

	sum_squares_f = (long double) sum_squares_hi *  pow (2,(sizeof(sum_squares)*8))  + sum_squares;
	mean = sum / num_samples;
	variance = (sum_squares_f - (((long double)sum * sum) / num_samples)) / num_samples;

	/* Debugging statistics */
#if DEBUG == 1
	printf ( "Debugging the statistics (values which won't fit long longs):\n"
		 "  Quadrant_0: sum=%15llu    sum_squares_hi=%2u    sum_squares=%21llu    Number=%8u\n"
		 "  Quadrant_1: sum=%15llu    sum_squares_hi=%2u    sum_squares=%21llu    Number=%8u\n"
		 "  Quadrant_2: sum=%15llu    sum_squares_hi=%2u    sum_squares=%21llu    Number=%8u\n"
		 "  Quadrant_3: sum=%15llu    sum_squares_hi=%2u    sum_squares=%21llu    Number=%8u\n"
		 "  Overall   : sum=%15llu    sum_squares_hi=%2u    sum_squares=%21llu    Number=%8u\n",
		sum_q[0],  sum_squares_hi_q[0], sum_squares_q[0], num_samples_q[0],
		sum_q[1],  sum_squares_hi_q[1], sum_squares_q[1], num_samples_q[1],
		sum_q[2],  sum_squares_hi_q[2], sum_squares_q[2], num_samples_q[2],
		sum_q[3],  sum_squares_hi_q[3], sum_squares_q[3], num_samples_q[3],
		sum,       sum_squares_hi,      sum_squares,      num_samples
		);

	printf ( "Detailed statistics:\n"
		 "  Quadrant_0:  Mean=%12.2Lf    Variance=%18.2Lf    StdDev=%12.2lf    Number=%8u\n"
		 "  Quadrant_1:  Mean=%12.2Lf    Variance=%18.2Lf    StdDev=%12.2lf    Number=%8u\n"
		 "  Quadrant_2:  Mean=%12.2Lf    Variance=%18.2Lf    StdDev=%12.2lf    Number=%8u\n"
		 "  Quadrant_3:  Mean=%12.2Lf    Variance=%18.2Lf    StdDev=%12.2lf    Number=%8u\n"
		 "  Overall   :  Mean=%12.2Lf    Variance=%18.2Lf    StdDev=%12.2lf    Number=%8u\n",
		 mean_q[0], variance_q[0], sqrt(fabs(variance_q[0])), num_samples_q[0],
		 mean_q[1], variance_q[1], sqrt(fabs(variance_q[1])), num_samples_q[1],
		 mean_q[2], variance_q[2], sqrt(fabs(variance_q[2])), num_samples_q[2],
		 mean_q[3], variance_q[3], sqrt(fabs(variance_q[3])), num_samples_q[3],
		 mean, 	    variance,      sqrt(fabs(variance)),      num_samples
		);
#endif

	/* Print statistical summary */
	printf ( "Pixel statistics for %s:\n"
		 "  Quadrant_0:  Mean=%12.2Lf    StdDev=%12.2lf    Number=%8u\n"
		 "  Quadrant_1:  Mean=%12.2Lf    StdDev=%12.2lf    Number=%8u\n"
		 "  Quadrant_2:  Mean=%12.2Lf    StdDev=%12.2lf    Number=%8u\n"
		 "  Quadrant_3:  Mean=%12.2Lf    StdDev=%12.2lf    Number=%8u\n"
		 "  Overall   :  Mean=%12.2Lf    StdDev=%12.2lf    Number=%8u\n",
		 cam_filename,
		 mean_q[0], sqrt(fabs(variance_q[0])), num_samples_q[0],
		 mean_q[1], sqrt(fabs(variance_q[1])), num_samples_q[1],
		 mean_q[2], sqrt(fabs(variance_q[2])), num_samples_q[2],
		 mean_q[3], sqrt(fabs(variance_q[3])), num_samples_q[3],
		 mean, 	    sqrt(fabs(variance)),      num_samples
		);

	if (errmsg_once == 0){   /* normally, return zero. If a fatal error occurred, we already exited (1). If a non-fatal, but serious error occurred, exit(2). */
		return (0);
	}else{
		return (2);
	}
}
