#!/bin/bash
#Read a single pixel value out of a .cam or .tiff file, and print it to stdout.
#Copyright Richard Neill, 2011. Free Software, Licensed under GNU GPL v3+


NUM_PXQ=512			#Pixels per quad
DEBUG=false			#Debug: set true for on, false for off.

NUM_PXA=$((NUM_PXQ * 2))	#Pixels per array side
max_pxq=$((NUM_PXQ - 1))	#max pixel value in a quad
max_pxa=$((NUM_PXA -1))		#max pixel value in array.


#Check binaries
stream=`which stream`;
if [ $? != 0 ];then
	echo "Error: program 'stream' not found. This is part of the ImageMagick package, configured --with-quantum-depth=32" >&2
	exit 1
fi
xxd=`which xxd`;
if [ $? != 0 ];then
	echo "Error: program 'xxd' not found." >&2
	exit 1
fi

#Check that stream has been built with 32-bit quantum depth.
stream_depth=`$stream --version | head -n1 | grep -oE ' Q(32|16) '| grep -oE '[0-9]*' `	#Version: ImageMagick 6.6.1-10 2011-03-28 Q32 http://www.imagemagick.org

USAGE () {
	echo ""
	echo "getpixel: extract the numerical value of a single pixel from a .cam or .tiff file"
	echo ""
	echo "USAGE:      getpixel -q QUAD -x X_POS -y Y_POS IMG_FILE"
	echo "OPTIONS:    -q QUAD     is the quadrant number:  0,1,2,3, or F, or X11"
	echo "            -x X_POS    is the x-co-ordinate, measured from the left. Range: [0, $max_pxq]."
	echo "            -y Y_POS    is the y-co-ordinate, measured from the bottom. Range: [0, $max_pxq]."
	echo ""
	echo "            -c, -r      are accepted as synonyms for -x, -y"
	echo "            -h          shows help"
	echo ""
	echo "IMG_FILE is either a .cam (.dat) or .tiff (.tif) file: either the cam or tiff"
	echo "format will be selected, depending on the extension."
	echo ""
	echo "MODES:      QUADRANT    (-q 0,1,2,3):  x,y are the Hawaii-coordinates, within that quadrant"
	echo "            FULL-FRAME  (-q F):        x,y are the Hawaii-coordinates, full-chip. Range: [0, $max_pxa]."
	echo "            X11-mode    (-q X11):      x,y are the X11-coordinates: full-frame, origin in top left."
	echo ""
	echo "See 'cam2tiff -h' for a diagram of the Hawaii chip, quadrants, and coordinate-axes."
	echo "Synonyms: x = column/fast-scan/pixel_clock; y = row/slow-scan/line_clock."
	echo "The output is a single 24-bit value, in the format 0xffeedd."
	echo ""
	echo "NOTE:  'stream' must be built with 32-bit quantum depth. This system has $stream_depth."
	echo "       To set this, build ImageMagick, configured --with-quantum-depth=32 ."
	echo ""
}

#Support --help, but not other longopts
if [ "$1" == --help ];then
	USAGE;
	exit 0
fi

#Getopts
while getopts c:r:q:x:y:h OPTION
do
        case "$OPTION" in
                c) X=$OPTARG;;
                r) Y=$OPTARG;;
                q) Q=$OPTARG;;
                x) X=$OPTARG;;
                y) Y=$OPTARG;;
                h|\?) USAGE ; exit 1 ;;
        esac
done
shift $((OPTIND-1))
FILE=$1
EXTRA=$2

#Check stream quantum depth. (Do this now, after -h, so that the manpage can always be generated).
if [ "$stream_depth" != 32 ];then
	echo "Error: 'stream' (i.e. '$stream') wasn't compiled with sufficient quantum-depth (has Q$stream_depth; need Q32). Rebuild ImageMagick, configured --with-quantum-depth=32" >&2
	exit 1
fi

#Sanity check: all args required.
if [ $# == 0 ];then
	USAGE;
	exit 1
fi
if [ -z "$X" ];then
	echo "Error: x-coordinate (-x or -c) is unset. This option is required." >&2
	exit 1
fi
if [ -z "$Y" ];then
	echo "Error: y-coordinate (-y or -r) is unset. This option is required." >&2
	exit 1
fi
if [ -z "$X" ];then
	echo "Error: quadrant (-q) is unset. This option is required." >&2
	exit 1
fi
if [ -z "$FILE" ];then
	echo "Error: file is unset. This argument is required." >&2
	exit 1
fi
if [ -n "$EXTRA" ];then
	echo "Error: too many arguments. Use -h for help." >&2
	exit 1
fi


#Sanity check: everything in range, file exists.
if [ ! -f "$FILE" ]; then
	echo "Error: data file '$FILE' does not exist." >&2
	exit 1
fi
extn=${FILE##*.}
if [ "$extn" == tiff -o "$extn" == tif ];then
	format=tiff
elif [ "$extn" == cam -o "$extn" == dat ];then
	format=cam
else
	echo "Error: data file '$FILE' has extension '$extn' which is not one of .cam, .tiff, .dat, .tif." >&2
	exit 1
fi
if [ "$Q" = F ];then		#Fullframe mode: equivalent to an expanded quadrant #1.
	fullframe=true
	max_pxq=$max_pxa
elif [ "$Q" = X11 ];then	#Fullframe mode, in X11 coordinates. invert y-axis.
	fullframe=true
	max_pxq=$max_pxa
	Q=F
	Y=$((max_pxa - Y))
	x11_txt="(X11 origin)"
elif [ "$Q" == 0 -o "$Q" == 1 -o "$Q" == 2 -o "$Q" == 3 ];then
	fullframe=false
else
	echo "Error: quadrant (-q) must be one of 0...3, or 'F' for full-frame, or 'X11' for full-frame X11-mode." >&2
	exit 1
fi
if [ "$X" -lt 0 -o "$X" -gt $max_pxq ];then
	echo "Error: x-coordinate must be within range 0...$max_pxq." >&2
	exit 1
fi
if [ "$Y" -lt 0 -o "$Y" -gt $max_pxq ];then
	echo "Error: y-coordinate must be within range 0...$max_pxq." >&2
	exit 1
fi

#Now validated.
[ $DEBUG == true ] && echo "Quad $Q, coordinates: X=$X, Y=$Y. File is $FILE, with extn $extn, and format $format. Fullframe = $fullframe $x11_txt" >&2


#Do it...
if [ $format == tiff ];then
	#Calculate pixel positions within the full-frame image. (converting to X11 coordinates: right/down)
					#Q	X	Y
	if   [ $Q == 0 ];then
		X=$((X + NUM_PXQ ))	#0	x+512   1023-y
		Y=$((max_pxa - Y))

	elif [ $Q == 1 ];then		#1	x	1023-y
		Y=$((max_pxa - Y))

	elif [ $Q == 2 ];then		#2	x	511-y
		Y=$((max_pxq - Y))

	elif [ $Q == 3 ];then		#3	x+512	511-y
		X=$((X + NUM_PXQ ))
		Y=$((max_pxq - Y))

	elif [ $Q == 'F' ];then		#Q	x	1023-y
		Y=$((max_pxa - Y))

	else
		echo "Error: invalid Q: $Q"
		exit 1
	fi
	[ $DEBUG == true ] && echo "Computed x,y (in X11 coordinates, for stream): x=$X, y=$Y." >&2

	#Extract the pixel value using stream. (Get a uint32, little endian. .tiff files have already been byte-shifted left, so revert this)
	[ $DEBUG == true ] && echo "Command: $stream -storage-type integer -colorspace gray -depth 32 -map i -extract 1x1+$X+$Y $FILE - | xxd -p" >&2
	pixel_value_uint32le=`$stream -storage-type integer -colorspace gray -depth 32 -map i -extract 1x1+$X+$Y $FILE - | xxd -p`

	pixel_value_hex="0x"${pixel_value_uint32le:6:2}${pixel_value_uint32le:4:2}${pixel_value_uint32le:2:2}
	[ $DEBUG == true ] && echo "Pixel value is $pixel_value_hex (extracted as uint32le: $pixel_value_uint32le)." >&2

else 	#cam

	#If fullframe, calculate the quadrant, and coordinates within quadrant.
	if [ $fullframe == true ];then
		if    [ $X -lt $NUM_PXQ -a $Y -lt $NUM_PXQ ] ;then
			Q=1

		elif  [ $X -lt $NUM_PXQ -a $Y -ge $NUM_PXQ ] ;then
			Q=2
			Y=$((Y - NUM_PXQ))

		elif  [ $X -ge $NUM_PXQ -a $Y -lt $NUM_PXQ ] ;then
			Q=0
			X=$((X - NUM_PXQ))

		elif  [ $X -ge $NUM_PXQ -a $Y -ge $NUM_PXQ ] ;then
			Q=3
			X=$((X - NUM_PXQ))
			Y=$((Y - NUM_PXQ))
		fi
	fi

	#Calculate the word-offset within the file.
	skip_words=$((($X +  $NUM_PXQ * $Y)* 4 + $Q))	# x + (512 * y) is raster, *4 because of interleaving, then add quadrant as offset
	skip_bytes=$((skip_words * 4))			#Wordlength=4 bytes.

	[ $DEBUG == true ] && echo "Computed skip for dd: words = $skip_words (bytes = $skip_bytes)." >&2

	#Extract the pixel value, using dd. (Get a uint24, little endian. No byte-shift to reverse)
	[ $DEBUG == true ] && echo "Command: dd if=$FILE bs=1 skip=$skip_bytes count=4 | xxd -p" >&2
	pixel_value_uint24le=`dd if=$FILE bs=1 skip=$skip_bytes count=4 2>/dev/null | xxd -p`

	pixel_value_hex="0x"${pixel_value_uint24le:4:2}${pixel_value_uint24le:2:2}${pixel_value_uint24le:0:2}
	[ $DEBUG == true ] && echo "Pixel value is $pixel_value_hex (extracted as uint24le: $pixel_value_uint24le)." >&2

fi

#Quick sanity check on result.
if echo "$pixel_value_hex" | grep -oqE '^0x[0-9a-f]{6}$' ;then
	echo "$pixel_value_hex"
	exit 0
else
	echo error
	echo "Error: extracted value '$pixel_value_hex' isn't sensible"
	exit 1
fi
